@@ -22,47 +22,50 @@
  * Copyright (C) 2002, Olaf Kirch <okir@lst.de>
  */
 
-#include <config.h>
-#include <string.h>
 #include "libp11-int.h"
+#include <string.h>
 
 static int pkcs11_find_certs(PKCS11_TOKEN *);
 static int pkcs11_next_cert(PKCS11_CTX *, PKCS11_TOKEN *, CK_SESSION_HANDLE);
-static int pkcs11_init_cert(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
-			    CK_SESSION_HANDLE session, CK_OBJECT_HANDLE o,
-			    PKCS11_CERT **);
-
-static CK_OBJECT_CLASS cert_search_class;
-static CK_ATTRIBUTE cert_search_attrs[] = {
-	{CKA_CLASS, &cert_search_class, sizeof(cert_search_class)},
-};
-#define numof(arr)	(sizeof(arr)/sizeof((arr)[0]))
+static int pkcs11_init_cert(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
+	CK_SESSION_HANDLE session, CK_OBJECT_HANDLE o, PKCS11_CERT **);
 
 /*
  * Enumerate all certs on the card
  */
-int
-PKCS11_enumerate_certs(PKCS11_TOKEN * token,
-		       PKCS11_CERT ** certp, unsigned int *countp)
+int pkcs11_enumerate_certs(PKCS11_TOKEN *token,
+		PKCS11_CERT **certp, unsigned int *countp)
 {
-	PKCS11_TOKEN_private *priv = PRIVTOKEN(token);
+	PKCS11_SLOT *slot = TOKEN2SLOT(token);
+	PKCS11_CTX *ctx = SLOT2CTX(slot);
+	PKCS11_TOKEN_private *tpriv = PRIVTOKEN(token);
+	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
+	int rv;
 
-	if (priv->ncerts < 0) {
-		priv->ncerts = 0;
-		if (pkcs11_find_certs(token)) {
-			pkcs11_destroy_certs(token);
-			return -1;
-		}
+	/* Make sure we have a session */
+	if (!spriv->haveSession && PKCS11_open_session(slot, 0))
+		return -1;
+
+	CRYPTO_THREAD_write_lock(cpriv->rwlock);
+	rv = pkcs11_find_certs(token);
+	CRYPTO_THREAD_unlock(cpriv->rwlock);
+	if (rv < 0) {
+		pkcs11_destroy_certs(token);
+		return -1;
 	}
-	*certp = priv->certs;
-	*countp = priv->ncerts;
+
+	if (certp)
+		*certp = tpriv->certs;
+	if (countp)
+		*countp = tpriv->ncerts;
 	return 0;
 }
 
 /*
  * Find certificate matching a key
  */
-PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY * key)
+PKCS11_CERT *pkcs11_find_certificate(PKCS11_KEY *key)
 {
 	PKCS11_KEY_private *kpriv;
 	PKCS11_CERT_private *cpriv;
@@ -75,7 +78,7 @@ PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY * key)
 	for (n = 0; n < count; n++, cert++) {
 		cpriv = PRIVCERT(cert);
 		if (cpriv->id_len == kpriv->id_len
-		    && !memcmp(cpriv->id, kpriv->id, kpriv->id_len))
+				&& !memcmp(cpriv->id, kpriv->id, kpriv->id_len))
 			return cert;
 	}
 	return NULL;
@@ -84,33 +87,32 @@ PKCS11_CERT *PKCS11_find_certificate(PKCS11_KEY * key)
 /*
  * Find all certs of a given type (public or private)
  */
-static int pkcs11_find_certs(PKCS11_TOKEN * token)
+static int pkcs11_find_certs(PKCS11_TOKEN *token)
 {
 	PKCS11_SLOT *slot = TOKEN2SLOT(token);
-	PKCS11_CTX *ctx = TOKEN2CTX(token);
-	CK_SESSION_HANDLE session;
+	PKCS11_CTX *ctx = SLOT2CTX(slot);
+	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
+	CK_OBJECT_CLASS cert_search_class;
+	CK_ATTRIBUTE cert_search_attrs[] = {
+		{CKA_CLASS, &cert_search_class, sizeof(cert_search_class)},
+	};
 	int rv, res = -1;
 
-	/* Make sure we have a session */
-	if (!PRIVSLOT(slot)->haveSession && PKCS11_open_session(slot, 0))
-		return -1;
-	session = PRIVSLOT(slot)->session;
-
 	/* Tell the PKCS11 lib to enumerate all matching objects */
 	cert_search_class = CKO_CERTIFICATE;
-	rv = CRYPTOKI_call(ctx, C_FindObjectsInit(session, cert_search_attrs,
-						  numof(cert_search_attrs)));
+	rv = CRYPTOKI_call(ctx, C_FindObjectsInit(spriv->session, cert_search_attrs, 1));
 	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_CERTS, rv);
 
 	do {
-		res = pkcs11_next_cert(ctx, token, session);
+		res = pkcs11_next_cert(ctx, token, spriv->session);
 	} while (res == 0);
 
-	CRYPTOKI_call(ctx, C_FindObjectsFinal(session));
+	CRYPTOKI_call(ctx, C_FindObjectsFinal(spriv->session));
+
 	return (res < 0) ? -1 : 0;
 }
 
-static int pkcs11_next_cert(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
+static int pkcs11_next_cert(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 		CK_SESSION_HANDLE session)
 {
 	CK_OBJECT_HANDLE obj;
@@ -130,101 +132,106 @@ static int pkcs11_next_cert(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
 	return 0;
 }
 
-static int pkcs11_init_cert(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
-	 CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj, PKCS11_CERT ** ret)
+static int pkcs11_init_cert(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
+		CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj, PKCS11_CERT ** ret)
 {
 	PKCS11_TOKEN_private *tpriv;
-	PKCS11_CERT_private *kpriv;
+	PKCS11_CERT_private *cpriv;
 	PKCS11_CERT *cert, *tmp;
-	char label[256], data[2048];
-	unsigned char id[256];
+	unsigned char *data;
 	CK_CERTIFICATE_TYPE cert_type;
 	size_t size;
+	int i;
 
-	size = sizeof(cert_type);
-	if (pkcs11_getattr_var(token, obj, CKA_CERTIFICATE_TYPE, &cert_type, &size))
+	(void)ctx;
+	(void)session;
+
+	/* Ignore unknown certificate types */
+	size = sizeof(CK_CERTIFICATE_TYPE);
+	if (pkcs11_getattr_var(token, obj, CKA_CERTIFICATE_TYPE, (CK_BYTE *)&cert_type, &size))
 		return -1;
-
-	/* Ignore any certs we don't understand */
 	if (cert_type != CKC_X_509)
 		return 0;
 
+	/* Prevent re-adding existing PKCS#11 object handles */
+	/* TODO: Rewrite the O(n) algorithm as O(log n),
+	 * or it may be too slow with a large number of certificates */
+	for (i=0; i < PRIVTOKEN(token)->ncerts; ++i)
+		if (PRIVCERT(PRIVTOKEN(token)->certs + i)->object == obj)
+			return 0;
+
+	/* Allocate memory */
+	cpriv = OPENSSL_malloc(sizeof(PKCS11_CERT_private));
+	if (cpriv == NULL)
+		return -1;
+	memset(cpriv, 0, sizeof(PKCS11_CERT_private));
 	tpriv = PRIVTOKEN(token);
-	tmp = (PKCS11_CERT *) OPENSSL_realloc(tpriv->certs,
-				(tpriv->ncerts + 1) * sizeof(PKCS11_CERT));
-	if (!tmp) {
-		free(tpriv->certs);
-		tpriv->certs = NULL;
+	tmp = OPENSSL_realloc(tpriv->certs,
+		(tpriv->ncerts + 1) * sizeof(PKCS11_CERT));
+	if (tmp == NULL)
 		return -1;
-	}
 	tpriv->certs = tmp;
-
 	cert = tpriv->certs + tpriv->ncerts++;
-	memset(cert, 0, sizeof(*cert));
-	cert->_private = kpriv = PKCS11_NEW(PKCS11_CERT_private);
-	kpriv->object = obj;
-	kpriv->parent = token;
+	memset(cert, 0, sizeof(PKCS11_CERT));
 
-	if (!pkcs11_getattr_s(token, obj, CKA_LABEL, label, sizeof(label)))
-		cert->label = BUF_strdup(label);
-	size = sizeof(data);
-	if (!pkcs11_getattr_var(token, obj, CKA_VALUE, data, &size)) {
-		const unsigned char *p = (unsigned char *) data;
+	/* Fill public properties */
+	pkcs11_getattr_alloc(token, obj, CKA_LABEL, (CK_BYTE **)&cert->label, NULL);
+	size = 0;
+	if (!pkcs11_getattr_alloc(token, obj, CKA_VALUE, &data, &size)) {
+		const unsigned char *p = data;
 
-		cert->x509 = d2i_X509(NULL, &p, size);
-	}
-	cert->id_len = sizeof(id);
-	if (!pkcs11_getattr_var(token, obj, CKA_ID, id, &cert->id_len)) {
-		cert->id = (unsigned char *) malloc(cert->id_len);
-		memcpy(cert->id, id, cert->id_len);
+		cert->x509 = d2i_X509(NULL, &p, (long)size);
+		OPENSSL_free(data);
 	}
+	cert->id_len = 0;
+	pkcs11_getattr_alloc(token, obj, CKA_ID, &cert->id, &cert->id_len);
 
-	/* Initialize internal information */
-	kpriv->id_len = sizeof(kpriv->id);
-	if (pkcs11_getattr_var(token, obj, CKA_ID, kpriv->id, &kpriv->id_len))
-		kpriv->id_len = 0;
+	/* Fill private properties */
+	cert->_private = cpriv;
+	cpriv->object = obj;
+	cpriv->parent = token;
+	cpriv->id_len = sizeof cpriv->id;
+	if (pkcs11_getattr_var(token, obj, CKA_ID, cpriv->id, &cpriv->id_len))
+		cpriv->id_len = 0;
 
 	if (ret)
 		*ret = cert;
-
 	return 0;
 }
 
 /*
  * Destroy all certs
  */
-void pkcs11_destroy_certs(PKCS11_TOKEN * token)
+void pkcs11_destroy_certs(PKCS11_TOKEN *token)
 {
-	PKCS11_TOKEN_private *priv = PRIVTOKEN(token);
+	PKCS11_TOKEN_private *tpriv = PRIVTOKEN(token);
 
-	while (priv->ncerts > 0) {
-		PKCS11_CERT *cert = &priv->certs[--(priv->ncerts)];
+	while (tpriv->ncerts > 0) {
+		PKCS11_CERT *cert = &tpriv->certs[--(tpriv->ncerts)];
 
 		if (cert->x509)
 			X509_free(cert->x509);
 		OPENSSL_free(cert->label);
 		if (cert->id)
-			free(cert->id);
+			OPENSSL_free(cert->id);
 		if (cert->_private != NULL)
 			OPENSSL_free(cert->_private);
 	}
-	if (priv->certs)
-		OPENSSL_free(priv->certs);
-	priv->ncerts = -1;
-	priv->certs = NULL;
+	if (tpriv->certs)
+		OPENSSL_free(tpriv->certs);
+	tpriv->certs = NULL;
+	tpriv->ncerts = 0;
 }
 
 /*
  * Store certificate
  */
-int
-PKCS11_store_certificate(PKCS11_TOKEN * token, X509 * x509, char *label,
-			 unsigned char *id, size_t id_len,
-			 PKCS11_CERT ** ret_cert)
+int pkcs11_store_certificate(PKCS11_TOKEN *token, X509 *x509, char *label,
+		unsigned char *id, size_t id_len, PKCS11_CERT ** ret_cert)
 {
 	PKCS11_SLOT *slot = TOKEN2SLOT(token);
-	PKCS11_CTX *ctx = TOKEN2CTX(token);
-	CK_SESSION_HANDLE session;
+	PKCS11_CTX *ctx = SLOT2CTX(slot);
+	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
 	CK_OBJECT_HANDLE object;
 	CK_ATTRIBUTE attrs[32];
 	unsigned int n = 0;
@@ -233,7 +240,6 @@ PKCS11_store_certificate(PKCS11_TOKEN * token, X509 * x509, char *label,
 	/* First, make sure we have a session */
 	if (!PRIVSLOT(slot)->haveSession && PKCS11_open_session(slot, 1))
 		return -1;
-	session = PRIVSLOT(slot)->session;
 
 	/* Now build the template */
 	pkcs11_addattr_int(attrs + n++, CKA_CLASS, CKO_CERTIFICATE);
@@ -246,7 +252,7 @@ PKCS11_store_certificate(PKCS11_TOKEN * token, X509 * x509, char *label,
 		pkcs11_addattr(attrs + n++, CKA_ID, id, id_len);
 
 	/* Now call the pkcs11 module to create the object */
-	rv = CRYPTOKI_call(ctx, C_CreateObject(session, attrs, n, &object));
+	rv = CRYPTOKI_call(ctx, C_CreateObject(spriv->session, attrs, n, &object));
 
 	/* Zap all memory allocated when building the template */
 	pkcs11_zap_attrs(attrs, n);
@@ -254,5 +260,7 @@ PKCS11_store_certificate(PKCS11_TOKEN * token, X509 * x509, char *label,
 	CRYPTOKI_checkerr(PKCS11_F_PKCS11_STORE_CERTIFICATE, rv);
 
 	/* Gobble the key object */
-	return pkcs11_init_cert(ctx, token, session, object, ret_cert);
+	return pkcs11_init_cert(ctx, token, spriv->session, object, ret_cert);
 }
+
+/* vim: set noexpandtab: */
