@@ -19,21 +19,37 @@
 
 #include "libp11-int.h"
 #include <string.h>
+#include <openssl/ui.h>
 #include <openssl/bn.h>
 
 #ifdef _WIN32
 #define strncasecmp strnicmp
 #endif
 
+/* The maximum length of PIN */
+#define MAX_PIN_LENGTH   32
+
 static int pkcs11_find_keys(PKCS11_TOKEN *, unsigned int);
-static int pkcs11_next_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
+static int pkcs11_next_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 	CK_SESSION_HANDLE session, CK_OBJECT_CLASS type);
-static int pkcs11_init_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
+static int pkcs11_init_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 	CK_SESSION_HANDLE session, CK_OBJECT_HANDLE o,
 	CK_OBJECT_CLASS type, PKCS11_KEY **);
 static int pkcs11_store_key(PKCS11_TOKEN *, EVP_PKEY *, unsigned int,
 	char *, unsigned char *, size_t, PKCS11_KEY **);
 
+/* Set UI method to allow retrieving CKU_CONTEXT_SPECIFIC PINs interactively */
+int pkcs11_set_ui_method(PKCS11_CTX *ctx,
+		UI_METHOD *ui_method, void *ui_user_data)
+{
+	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
+	if (cpriv == NULL)
+		return -1;
+	cpriv->ui_method = ui_method;
+	cpriv->ui_user_data = ui_user_data;
+	return 0;
+}
+
 /*
  * Find key matching a certificate
  */
@@ -59,7 +75,7 @@ PKCS11_KEY *pkcs11_find_key(PKCS11_CERT *cert)
 /*
  * Find key matching a key of the other type (public vs private)
  */
-PKCS11_KEY *pkcs11_find_key_from_key(PKCS11_KEY * keyin)
+PKCS11_KEY *pkcs11_find_key_from_key(PKCS11_KEY *keyin)
 {
 	PKCS11_KEY_private *kinpriv = PRIVKEY(keyin);
 	PKCS11_KEY *keys;
@@ -80,7 +96,7 @@ PKCS11_KEY *pkcs11_find_key_from_key(PKCS11_KEY * keyin)
 /*
  * Reopens the object associated with the key
  */
-int pkcs11_reload_key(PKCS11_KEY * key)
+int pkcs11_reload_key(PKCS11_KEY *key)
 {
 	PKCS11_KEY_private *kpriv = PRIVKEY(key);
 	PKCS11_SLOT *slot = KEY2SLOT(key);
@@ -115,16 +131,16 @@ int pkcs11_reload_key(PKCS11_KEY * key)
  * FIXME: We should check first whether the token supports
  * on-board key generation, and if it does, use its own algorithm
  */
-int pkcs11_generate_key(PKCS11_TOKEN * token, int algorithm, unsigned int bits,
+int pkcs11_generate_key(PKCS11_TOKEN *token, int algorithm, unsigned int bits,
 		char *label, unsigned char* id, size_t id_len)
 {
 	PKCS11_KEY *key_obj;
 	EVP_PKEY *pk;
 	RSA *rsa;
 	BIO *err;
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-	BIGNUM *exp = NULL;
-	BN_GENCB *gencb = NULL;
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
+	BIGNUM *exp;
+	BN_GENCB *gencb;
 #endif
 	int rc;
 
@@ -135,20 +151,20 @@ int pkcs11_generate_key(PKCS11_TOKEN * token, int algorithm, unsigned int bits,
 
 	err = BIO_new_fp(stderr, BIO_NOCLOSE);
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 	exp = BN_new();
 	rsa = RSA_new();
 	gencb = BN_GENCB_new();
 	if (gencb)
-	    BN_GENCB_set(gencb, NULL, err);
-
-	if ( rsa == NULL  || exp == NULL || gencb == NULL
-	    || !BN_set_word(exp, RSA_F4) || !RSA_generate_key_ex(rsa, bits, exp, gencb)) {
+		BN_GENCB_set(gencb, NULL, err);
+	if (rsa == NULL || exp == NULL || gencb == NULL ||
+			!BN_set_word(exp, RSA_F4) ||
+			!RSA_generate_key_ex(rsa, bits, exp, gencb)) {
 		RSA_free(rsa);
+		rsa = NULL;
 	}
 	BN_GENCB_free(gencb);
 	BN_free(exp);
-
 #else
 	rsa = RSA_generate_key(bits, RSA_F4, NULL, err);
 #endif
@@ -177,7 +193,7 @@ int pkcs11_generate_key(PKCS11_TOKEN * token, int algorithm, unsigned int bits,
 /*
  * Store a private key on the token
  */
-int pkcs11_store_private_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
+int pkcs11_store_private_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 		char *label, unsigned char *id, size_t id_len)
 {
 	if (pkcs11_store_key(token, pk, CKO_PRIVATE_KEY, label, id, id_len, NULL))
@@ -185,7 +201,7 @@ int pkcs11_store_private_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 	return 0;
 }
 
-int pkcs11_store_public_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
+int pkcs11_store_public_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 		char *label, unsigned char *id, size_t id_len)
 {
 	if (pkcs11_store_key(token, pk, CKO_PUBLIC_KEY, label, id, id_len, NULL))
@@ -196,7 +212,7 @@ int pkcs11_store_public_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 /*
  * Store private key
  */
-static int pkcs11_store_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
+static int pkcs11_store_key(PKCS11_TOKEN *token, EVP_PKEY *pk,
 		unsigned int type, char *label, unsigned char *id, size_t id_len,
 		PKCS11_KEY ** ret_key)
 {
@@ -207,7 +223,7 @@ static int pkcs11_store_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 	CK_ATTRIBUTE attrs[32];
 	unsigned int n = 0;
 	int rv;
-	BIGNUM *rsa_n, *rsa_e, *rsa_d, *rsa_p, *rsa_q;
+	const BIGNUM *rsa_n, *rsa_e, *rsa_d, *rsa_p, *rsa_q;
 
 	/* First, make sure we have a session */
 	if (!spriv->haveSession && PKCS11_open_session(slot, 1))
@@ -231,7 +247,7 @@ static int pkcs11_store_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 		pkcs11_addattr_bool(attrs + n++, CKA_VERIFY, TRUE);
 		pkcs11_addattr_bool(attrs + n++, CKA_WRAP, TRUE);
 	}
-#if OPENSSL_VERSION_NUMBER >= 0x10100003L
+#if OPENSSL_VERSION_NUMBER >= 0x10100003L && !defined(LIBRESSL_VERSION_NUMBER)
 	if (EVP_PKEY_base_id(pk) == EVP_PKEY_RSA) {
 		RSA *rsa = EVP_PKEY_get1_RSA(pk);
 #else
@@ -239,7 +255,7 @@ static int pkcs11_store_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 		RSA *rsa = pk->pkey.rsa;
 #endif
 		pkcs11_addattr_int(attrs + n++, CKA_KEY_TYPE, CKK_RSA);
-#if OPENSSL_VERSION_NUMBER >= 0x10100005L
+#if OPENSSL_VERSION_NUMBER >= 0x10100005L && !defined(LIBRESSL_VERSION_NUMBER)
 		RSA_get0_key(rsa, &rsa_n, &rsa_e, &rsa_d);
 		RSA_get0_factors(rsa, &rsa_p, &rsa_q);
 #else
@@ -280,7 +296,7 @@ static int pkcs11_store_key(PKCS11_TOKEN * token, EVP_PKEY * pk,
 /*
  * Get the key type
  */
-int pkcs11_get_key_type(PKCS11_KEY * key)
+int pkcs11_get_key_type(PKCS11_KEY *key)
 {
 	PKCS11_KEY_private *kpriv = PRIVKEY(key);
 
@@ -309,7 +325,7 @@ EVP_PKEY *pkcs11_get_key(PKCS11_KEY *key, int isPrivate)
 				fprintf(stderr, "Missing CKA_ALWAYS_AUTHENTICATE attribute\n");
 		}
 	}
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L && !defined(LIBRESSL_VERSION_NUMBER)
 	EVP_PKEY_up_ref(key->evp_key);
 #else
 	CRYPTO_add(&key->evp_key->references, 1, CRYPTO_LOCK_EVP_PKEY);
@@ -322,29 +338,62 @@ EVP_PKEY *pkcs11_get_key(PKCS11_KEY *key, int isPrivate)
  */
 int pkcs11_authenticate(PKCS11_KEY *key)
 {
-	PKCS11_KEY_private *kpriv = PRIVKEY(key);
-	PKCS11_SLOT *slot = KEY2SLOT(key);
+	PKCS11_TOKEN *token = KEY2TOKEN(key);
+	PKCS11_SLOT *slot = TOKEN2SLOT(token);
 	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
 	PKCS11_CTX *ctx = SLOT2CTX(slot);
+	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
+	char pin[MAX_PIN_LENGTH+1];
+	char* prompt;
+	UI *ui;
 	int rv;
 
-	if (!kpriv->always_authenticate)
-		return 0;
+	/* Handle CKF_PROTECTED_AUTHENTICATION_PATH */
+	if (token->secureLogin) {
+		rv = CRYPTOKI_call(ctx,
+			C_Login(spriv->session, CKU_CONTEXT_SPECIFIC, NULL, 0));
+		return rv == CKR_USER_ALREADY_LOGGED_IN ? 0 : rv;
+	}
+
+	/* Call UI to ask for a PIN */
+	ui = UI_new_method(cpriv->ui_method);
+	if (ui == NULL)
+		return PKCS11_UI_FAILED;
+	if (cpriv->ui_user_data != NULL)
+		UI_add_user_data(ui, cpriv->ui_user_data);
+	memset(pin, 0, MAX_PIN_LENGTH+1);
+	prompt = UI_construct_prompt(ui, "PKCS#11 key PIN", key->label);
+	if (!prompt) {
+		return PKCS11_UI_FAILED;
+	}
+	if (!UI_dup_input_string(ui, prompt,
+			UI_INPUT_FLAG_DEFAULT_PWD, pin, 4, MAX_PIN_LENGTH)) {
+		UI_free(ui);
+		OPENSSL_free(prompt);
+		return PKCS11_UI_FAILED;
+	}
+	OPENSSL_free(prompt);
+
+	if (UI_process(ui)) {
+		UI_free(ui);
+		return PKCS11_UI_FAILED;
+	}
+	UI_free(ui);
+
+	/* Login with the PIN */
 	rv = CRYPTOKI_call(ctx,
 		C_Login(spriv->session, CKU_CONTEXT_SPECIFIC,
-			(CK_UTF8CHAR *)spriv->prev_pin,
-			spriv->prev_pin ? strlen(spriv->prev_pin) : 0));
-	if (rv == CKR_USER_ALREADY_LOGGED_IN) /* ignore */
-		rv = 0;
-	return rv;
+			(CK_UTF8CHAR *)pin, strlen(pin)));
+	OPENSSL_cleanse(pin, MAX_PIN_LENGTH+1);
+	return rv == CKR_USER_ALREADY_LOGGED_IN ? 0 : rv;
 }
 
 /*
  * Return keys of a given type (public or private)
  * Use the cached values if available
  */
-int pkcs11_enumerate_keys(PKCS11_TOKEN * token, unsigned int type,
-		PKCS11_KEY ** keyp, unsigned int * countp)
+int pkcs11_enumerate_keys(PKCS11_TOKEN *token, unsigned int type,
+		PKCS11_KEY ** keyp, unsigned int *countp)
 {
 	PKCS11_SLOT *slot = TOKEN2SLOT(token);
 	PKCS11_CTX *ctx = TOKEN2CTX(token);
@@ -352,20 +401,31 @@ int pkcs11_enumerate_keys(PKCS11_TOKEN * token, unsigned int type,
 	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
 	PKCS11_CTX_private *cpriv = PRIVCTX(ctx);
 	PKCS11_keys *keys = (type == CKO_PRIVATE_KEY) ? &tpriv->prv : &tpriv->pub;
+	PKCS11_KEY *first_key_prev = keys->keys;
 	int rv;
+	int i;
 
-	if (keys->num < 0) { /* No cache was built for the specified type */
-		/* Make sure we have a session */
-		if (!spriv->haveSession && PKCS11_open_session(slot, 0))
-			return -1;
-		CRYPTO_THREAD_write_lock(cpriv->rwlock);
-		rv = pkcs11_find_keys(token, type);
-		CRYPTO_THREAD_unlock(cpriv->rwlock);
-		if (rv < 0) {
-			pkcs11_destroy_keys(token, type);
-			return -1;
+	/* Make sure we have a session */
+	if (!spriv->haveSession && PKCS11_open_session(slot, 0))
+		return -1;
+
+	CRYPTO_THREAD_write_lock(cpriv->rwlock);
+	rv = pkcs11_find_keys(token, type);
+	CRYPTO_THREAD_unlock(cpriv->rwlock);
+	if (rv < 0) {
+		pkcs11_destroy_keys(token, type);
+		return -1;
+	}
+
+	/* Always update key references if the keys pointer changed */
+	if (first_key_prev != NULL && first_key_prev != keys->keys) {
+		for (i = 0; i < keys->num; ++i) {
+			PKCS11_KEY *key = keys->keys + i;
+			PKCS11_KEY_private *kpriv = PRIVKEY(key);
+			kpriv->ops->update_ex_data(key);
 		}
 	}
+
 	if (keyp)
 		*keyp = keys->keys;
 	if (countp)
@@ -376,13 +436,11 @@ int pkcs11_enumerate_keys(PKCS11_TOKEN * token, unsigned int type,
 /*
  * Find all keys of a given type (public or private)
  */
-static int pkcs11_find_keys(PKCS11_TOKEN * token, unsigned int type)
+static int pkcs11_find_keys(PKCS11_TOKEN *token, unsigned int type)
 {
 	PKCS11_SLOT *slot = TOKEN2SLOT(token);
 	PKCS11_CTX *ctx = TOKEN2CTX(token);
-	PKCS11_TOKEN_private *tpriv = PRIVTOKEN(token);
 	PKCS11_SLOT_private *spriv = PRIVSLOT(slot);
-	PKCS11_keys *keys = (type == CKO_PRIVATE_KEY) ? &tpriv->prv : &tpriv->pub;
 	CK_OBJECT_CLASS key_search_class;
 	CK_ATTRIBUTE key_search_attrs[1] = {
 		{CKA_CLASS, &key_search_class, sizeof(key_search_class)},
@@ -395,7 +453,6 @@ static int pkcs11_find_keys(PKCS11_TOKEN * token, unsigned int type)
 		C_FindObjectsInit(spriv->session, key_search_attrs, 1));
 	CRYPTOKI_checkerr(PKCS11_F_PKCS11_ENUM_KEYS, rv);
 
-	keys->num = 0;
 	do {
 		res = pkcs11_next_key(ctx, token, spriv->session, type);
 	} while (res == 0);
@@ -405,7 +462,7 @@ static int pkcs11_find_keys(PKCS11_TOKEN * token, unsigned int type)
 	return (res < 0) ? -1 : 0;
 }
 
-static int pkcs11_next_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
+static int pkcs11_next_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 		CK_SESSION_HANDLE session, CK_OBJECT_CLASS type)
 {
 	CK_OBJECT_HANDLE obj;
@@ -425,7 +482,7 @@ static int pkcs11_next_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
 	return 0;
 }
 
-static int pkcs11_init_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
+static int pkcs11_init_key(PKCS11_CTX *ctx, PKCS11_TOKEN *token,
 		CK_SESSION_HANDLE session, CK_OBJECT_HANDLE obj,
 		CK_OBJECT_CLASS type, PKCS11_KEY ** ret)
 {
@@ -436,14 +493,15 @@ static int pkcs11_init_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
 	CK_KEY_TYPE key_type;
 	PKCS11_KEY_ops *ops;
 	size_t size;
+	int i;
 
 	(void)ctx;
 	(void)session;
 
-	size = sizeof(key_type);
+	/* Ignore unknown key types */
+	size = sizeof(CK_KEY_TYPE);
 	if (pkcs11_getattr_var(token, obj, CKA_KEY_TYPE, (CK_BYTE *)&key_type, &size))
 		return -1;
-
 	switch (key_type) {
 	case CKK_RSA:
 		ops = &pkcs11_rsa_ops;
@@ -458,30 +516,36 @@ static int pkcs11_init_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
 		return 0;
 	}
 
-	tmp = OPENSSL_realloc(keys->keys, (keys->num + 1) * sizeof(PKCS11_KEY));
-	if (tmp == NULL) {
-		OPENSSL_free(keys->keys);
-		keys->keys = NULL;
-		return -1;
-	}
-	keys->keys = tmp;
+	/* Prevent re-adding existing PKCS#11 object handles */
+	/* TODO: Rewrite the O(n) algorithm as O(log n),
+	 * or it may be too slow with a large number of keys */
+	for (i=0; i < keys->num; ++i)
+		if (PRIVKEY(keys->keys + i)->object == obj)
+			return 0;
 
+	/* Allocate memory */
+	kpriv = OPENSSL_malloc(sizeof(PKCS11_KEY_private));
+	if (kpriv == NULL)
+		return -1;
+	memset(kpriv, 0, sizeof(PKCS11_KEY_private));
+	tmp = OPENSSL_realloc(keys->keys, (keys->num + 1) * sizeof(PKCS11_KEY));
+	if (tmp == NULL)
+		return -1;
+	keys->keys = tmp;
 	key = keys->keys + keys->num++;
 	memset(key, 0, sizeof(PKCS11_KEY));
-	kpriv = OPENSSL_malloc(sizeof(PKCS11_KEY_private));
-	if (kpriv)
-		memset(kpriv, 0, sizeof(PKCS11_KEY_private));
-	key->_private = kpriv;
-	kpriv->object = obj;
-	kpriv->parent = token;
 
+	/* Fill public properties */
 	pkcs11_getattr_alloc(token, obj, CKA_LABEL, (CK_BYTE **)&key->label, NULL);
 	key->id_len = 0;
 	pkcs11_getattr_alloc(token, obj, CKA_ID, &key->id, &key->id_len);
 	key->isPrivate = (type == CKO_PRIVATE_KEY);
 
-	/* Initialize internal information */
-	kpriv->id_len = sizeof(kpriv->id);
+	/* Fill private properties */
+	key->_private = kpriv;
+	kpriv->object = obj;
+	kpriv->parent = token;
+	kpriv->id_len = sizeof kpriv->id;
 	if (pkcs11_getattr_var(token, obj, CKA_ID, kpriv->id, &kpriv->id_len))
 		kpriv->id_len = 0;
 	kpriv->ops = ops;
@@ -495,7 +559,7 @@ static int pkcs11_init_key(PKCS11_CTX * ctx, PKCS11_TOKEN * token,
 /*
  * Destroy all keys of a given type (public or private)
  */
-void pkcs11_destroy_keys(PKCS11_TOKEN * token, unsigned int type)
+void pkcs11_destroy_keys(PKCS11_TOKEN *token, unsigned int type)
 {
 	PKCS11_TOKEN_private *tpriv = PRIVTOKEN(token);
 	PKCS11_keys *keys = (type == CKO_PRIVATE_KEY) ? &tpriv->prv : &tpriv->pub;
@@ -514,7 +578,7 @@ void pkcs11_destroy_keys(PKCS11_TOKEN * token, unsigned int type)
 	if (keys->keys)
 		OPENSSL_free(keys->keys);
 	keys->keys = NULL;
-	keys->num = -1;
+	keys->num = 0;
 }
 
 /* vim: set noexpandtab: */
